import device

import wiringpi as wp


class GPIODevice(device.Device):
    def __init__(self) -> None:
        super().__init__()
        wp.wiringPiSetupGpio()

    def update(self, id):
        if self.pins[id] == device.INPUT:
            self.data[id] = wp.digitalRead(id)
        else:
            wp.digitalWrite(id, self.data[id])

    def setup(self, id, mode):
        if mode == device.INPUT:
            wp.pinMode(id, 0)
        else:
            wp.pinMode(id, 1)
        return super().setup(id, mode)





class Block:
    def __init__(self, blockSize) -> None:
        self.blockSize = blockSize
        self.currentIndex = 0
        self.block = [0] * self.blockSize
        self.currentSum = 0

    def add(self, value):
        self.currentSum -= self.block[self.currentIndex]
        self.block[self.currentIndex] = value
        self.currentSum += self.block[self.currentIndex]
        self.currentIndex = (self.currentIndex + 1) % self.blockSize

    def get(self):
        return self.currentSum/self.blockSize


class Debouncer:
    def __init__(self, updatePeriod, addPeriod, blockSize) -> None:
        self.updatePeriod = updatePeriod
        self.lastUpdateTime = 0

        self.addPeriod = addPeriod
        self.lastAddTime = 0

        self.block = Block(blockSize)

    def get(self):
        return self.block.get()

    def updateValueIfNeed(self, currentTime, func):
        if currentTime - self.lastUpdateTime < self.updatePeriod:
            return False
        func()
        self.lastUpdateTime = currentTime
        return True

    def addToBufferIfNeed(self, currentTime, value):
        if currentTime - self.lastAddTime < self.addPeriod:
            return False
        self.block.add(value)
        self.lastAddTime = currentTime
        return True



from HyIo.debounce import Debouncer
from HyIo.device import Device
from HyIo.gpioDevice import GPIODevice
from HyIo.i2cDevice import I2CDevice
from HyIo.pin import Pin, InputPin, OutputPin
from HyIo.ioManager import IoManager



import pin
import device
import debounce


def test_pin():
    dev = device.Device()

    inp = pin.InputPin(1, debounce.Debouncer(0, 0, 1), dev)
    inp.setup()

    out = pin.OutputPin(2, debounce.Debouncer(0, 0, 1), dev)
    out.setup()

    inp.update(1)
    assert inp.get() == 0
    dev.set(inp.id, 1)
    inp.update(2)
    assert inp.get() == 0

    out.update(1)
    assert out.get() == 0
    out.set(1)
    out.update(2)
    assert out.get() == 1



import ioManager
import device
import pin
import debounce
import time


def test_manager():
    manager = ioManager.IoManager(10)

    dev = device.Device()

    out = pin.OutputPin(1, debounce.Debouncer(0, 0, 1), dev)
    out.setup()

    global currentValue
    currentValue = 0

    def notify_me(value):
        global currentValue
        currentValue = value
        if value > 10:
            currentValue = (15/0)

    manager.registerPin(out, notify_me)

    manager.startThread()
    time.sleep(0.02)
    out.set(1)
    time.sleep(0.02)
    assert currentValue == 1
    out.set(0)
    time.sleep(0.02)
    assert currentValue == 0

    assert manager.lastError == ""

    out.set(11)
    time.sleep(0.02)
    assert len(manager.lastError) > 0

    manager.stopThread()



import time
import threading
import queue
import traceback


class IoManager:
    def __init__(self, period) -> None:
        self.period = period
        self.pins = []
        self.callbacks = []
        self.queue = queue.Queue()
        self.active = True
        self.updateThreadObj = None
        self.callbackThreadObj = None
        self.threadActive = False
        self.lastError = ""
        self.locals = None
        self.globals = None

    def setLocals(self, local):
        self.locals = local

    def setGlobals(self, globa):
        self.globals = globa

    def registerPin(self, pin, callback):
        self.pins += [pin]
        self.callbacks += [callback]

    def updatePins(self):
        currentTime = time.perf_counter_ns()
        for i in range(len(self.pins)):
            if self.pins[i].update(currentTime):
                data = (self.callbacks[i], self.pins[i].get())
                self.queue.put(data, block=True)

    def _updateThreadFunction(self):
        while self.threadActive:
            self.updatePins()
            time.sleep(self.period/1000)

    def _callbackThreadFunction(self):
        while self.threadActive:
            if self.queue.empty():
                time.sleep(self.period/1000)
                continue
            func, args = self.queue.get(block=True)
            try:
                func(args)
            except BaseException:
                self.lastError = traceback.format_exc()

    def startThread(self):
        if self.threadActive:
            return
        self.threadActive = True
        self.callbackThreadObj = threading.Thread(target=self._callbackThreadFunction)
        self.updateThreadObj = threading.Thread(target=self._updateThreadFunction)

        self.callbackThreadObj.start()
        self.updateThreadObj.start()

    def stopThread(self):
        if not self.threadActive:
            return
        self.threadActive = False

        if self.updateThreadObj is not None:
            self.updateThreadObj.join()

        if self.callbackThreadObj is not None:
            self.callbackThreadObj.join()

        self.updateThreadObj = self.callbackThreadObj = None



import device


def test_device():
    dev = device.Device()

    dev.setup(1, device.INPUT)
    dev.setup(2, device.OUTPUT)
    dev.setup(3, device.INPUT)
    dev.setup(4, device.OUTPUT)

    for inp in [1, 3]:
        assert dev.get(inp) == 0
        dev.set(inp, 1)
        assert dev.get(inp) == 0
    for out in [2, 4]:
        assert dev.get(out) == 0
        dev.set(out, 1)
        assert dev.get(out) == 1



import device


class Pin:
    def __init__(self, id, debouncer, device) -> None:
        self.id = id
        self.debouncer = debouncer
        self.device = device
        self._oldValue = 0

    def setup(self, mode=device.NONE):
        self.device.setup(self.id, mode)

    def _updatePinOnDeviceFunction(self):
        self.device.update(self.id)

    def _setPinOnDeviceFunction(self):
        self.device.set(self.id, self.get())

    def _getFromDeviceFunction(self):
        return self.device.get(self.id)

    def update(self, currentTime):
        currentValue = self.get()
        if currentValue == self._oldValue:
            return False
        self._oldValue = currentValue
        return True

    def get(self):
        return self.debouncer.get()

    def set(self, value):
        pass


class InputPin(Pin):
    def __init__(self, id, debouncer, device) -> None:
        super().__init__(id, debouncer, device)

    def setup(self):
        return super().setup(device.INPUT)

    def update(self, currentTime):
        self.debouncer.updateValueIfNeed(currentTime, self._updatePinOnDeviceFunction)  # function is executed inside
        self.debouncer.addToBufferIfNeed(currentTime, self._getFromDeviceFunction())  # get executed here
        return super().update(currentTime)


class OutputPin(Pin):
    def __init__(self, id, debouncer, device) -> None:
        super().__init__(id, debouncer, device)
        self._lastOutValue = 0

    def setup(self):
        return super().setup(device.OUTPUT)

    def update(self, currentTime):
        self.debouncer.addToBufferIfNeed(currentTime, self._lastOutValue)
        self.debouncer.updateValueIfNeed(currentTime, self._setPinOnDeviceFunction)  # function is executed inside
        return super().update(currentTime)

    def set(self, value):
        self._lastOutValue = value



import debounce


def test_block():
    for window in range(1, 100):
        block = debounce.Block(window)
        for i in range(1, 1000):
            block.add(i)

            currentLen = min(i, window)
            currentSum = ((i - (currentLen-1)) + i)*currentLen/2
            avg = currentSum / window
            assert avg == block.get()


def test_debouncer():
    def functor(*args):
        pass
    debouncer = debounce.Debouncer(10, 10, 1)

    for curTime in range(0, 10):
        assert not debouncer.updateValueIfNeed(curTime, functor)
    assert debouncer.updateValueIfNeed(10, functor)
    for curTime in range(11, 20):
        assert not debouncer.updateValueIfNeed(curTime, functor)

    for curTime in range(0, 10):
        assert not debouncer.addToBufferIfNeed(curTime, curTime)
    assert debouncer.addToBufferIfNeed(10, 10)
    assert debouncer.get() == 10
    for curTime in range(11, 20):
        assert not debouncer.addToBufferIfNeed(curTime, curTime)
        assert debouncer.get() == 10



from debounce import Debouncer
from device import Device
from pin import Pin


# maybe this doesn't make sense. Since we should use callbacks,
# how do we specify a callback function inside a json. It could be the name of
# a external script, but we lose some context
# user can define a local/global vars, then set her to execute external scripts
"""
{
    "i2c-0": {
        "address": 0x20
    },
    "pin":{
        "id":0,
        "debouncer": {
            "updatePeriod": 0,
            "addPeriod": 0,
            "blockSize": 0
        },
        "device": "i2c-0",
        "device": "gpio",
        "device": "sim",
        "mode": "input"
    }
}
"""



import smbus2
import device

class I2CDevice(device.Device):
    def __init__(self, address) -> None:
        super().__init__()
        self.bus = smbus2.SMBusWrapper(1)
        self.address = address

    def _read_pin(self, id):
        return self.bus.read_byte_data(self.address, id)

    def _write_pin(self, id, value):
        self.bus.write_byte_data(self.address, id, value)

    def update(self, id):
        if self.pins[id] == device.INPUT:
            self.data[id] = self._read_pin(id)
        else:
            self._write_pin(id, self.data[id])



from distutils.core import setup
setup(
  name = 'HyIo',
  packages = ['HyIo'],
  version = '0.2.2',
  license='MIT',
  description = 'Abstraction to Io device on raspberry pi',
  author = 'Erik Cruz Morbach',
  author_email = 'morbacherik@gmail.com',
  url = 'https://github.com/Erik-Morbach/HyIo',
  download_url = 'https://github.com/Erik-Morbach/HyIo/archive/refs/tags/V0.2.2.tar.gz',
  keywords = ['io', 'raspberry', 'abstraction'],
  install_requires=[
          'wiringpi',
          'smbus2',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
  ],
)



INPUT = "input"
OUTPUT = "output"
NONE = "none"


class Device:
    def __init__(self) -> None:
        self.data = {}
        self.pins = {}

    def setup(self, id, mode):
        self.pins[id] = mode
        self.data[id] = 0

    def update(self, id):  # set self.data[id] to hardware
        pass

    def set(self, id, value):
        if self.pins[id] == OUTPUT:
            self.data[id] = value

    def get(self, id):
        return self.data[id]
